/*********************************************************************************
 * Copyright (c) 2010 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/
package eu.unicore.hila.grid.unicore6;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.UUID;
import java.util.concurrent.Callable;

import eu.unicore.hila.HiLA;
import eu.unicore.hila.Resource;
import eu.unicore.hila.common.grid.BaseThirdPartyTransfer;
import eu.unicore.hila.exceptions.HiLAException;
import eu.unicore.hila.exceptions.HiLALocationSyntaxException;
import eu.unicore.hila.grid.File;
import eu.unicore.hila.grid.Progressable;
import eu.unicore.hila.grid.TaskStatus;
import eu.unicore.hila.grid.ThirdPartyTransfer;

/**
 * @author bjoernh
 * 
 *         30.06.2010 13:56:25
 * 
 */
public class Unicore6RecursiveTransferTask extends BaseThirdPartyTransfer
	implements ThirdPartyTransfer, Callable<TaskStatus>, Progressable {

    final private Unicore6File source;
    final private Unicore6File destination;

    final private List<ThirdPartyTransfer> transfers;

    private boolean abort = false;
    private boolean allCreated = false;

    /**
     * @param unicore6File
     * @param _sourceStorageClient
     * @param _destination
     * @param _destinationStorageClient
     * @throws HiLALocationSyntaxException
     */
    public Unicore6RecursiveTransferTask(Unicore6File _source,
         Unicore6File _destination)
	    throws HiLAException {
	super(_source.getStorage().getLocation().getChildLocation("transfers")
		.getChildLocation(UUID.randomUUID().toString()));

	this.source = _source;
	this.destination = _destination;

	this.transfers = new ArrayList<ThirdPartyTransfer>();

	setCurrentState(TaskStatus.NEW);

      HiLA.getExecutor().submit(this);

	Unicore6TransfersCollection tc = (Unicore6TransfersCollection) this
		.getParent();
	if (tc != null) {
	    tc.addTask(this);
	}
    }

    /**
     * @see eu.unicore.hila.Task#abort()
     */
    @Override
    public void abort() throws HiLAException {
	abort = true;
	synchronized (this) {
	    if (!allCreated) {
		try {
		    wait();
		} catch (InterruptedException e) {

		}
	    }
	}
	for (ThirdPartyTransfer transfer : transfers) {
	    transfer.abort();
	}
    }

    /**
     * @see eu.unicore.hila.Task#getId()
     */
    @Override
    public String getId() throws HiLAException {
	return this.location.getName();
    }

    /**
     * @see eu.unicore.hila.Resource#getChildren()
     */
    @Override
    public List<Resource> getChildren() throws HiLAException {
	return Collections.singletonList((Resource) null);
    }

    /**
     * @see java.util.concurrent.Callable#call()
     */
    @Override
    public TaskStatus call() throws Exception {
	setCurrentState(TaskStatus.RUNNING);

	if (!destination.exists()) {
	    destination.mkdir();
	}

	synchronized (this) {
	    List<File> files = this.source.ls();
	    for (File file : files) {
		transfers.add(file.transfer(this.destination, false, true));
		if (abort) {
		    break;
		}
	    }
	    allCreated = true;
	    notify();
	}

	// TODO blocking on them should be cheaper
	while (TasksHelper.hasPendingTasks(transfers)) {
	    Thread.sleep(2000);
	}

	if (TasksHelper.allTasksSuccessful(transfers)) {
	    setCurrentState(TaskStatus.SUCCESSFUL);

	} else {
	    HiLAException e = new HiLAException(
		    "Not all pending transfers were successful.");
	    setCurrentState(TaskStatus.FAILED, e);
	    throw e;

	}
	return status();
    }

}
